package com.grovestreams.sensor.gs_gmetad;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.zip.GZIPOutputStream;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.Consts;
import org.apache.http.HttpEntity;
import org.apache.http.HttpException;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.NameValuePair;
import org.apache.http.client.methods.HttpDelete;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.methods.HttpPut;
import org.apache.http.client.methods.HttpRequestBase;
import org.apache.http.client.utils.URLEncodedUtils;
import org.apache.http.entity.ByteArrayEntity;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.params.CoreConnectionPNames;
import org.apache.http.util.EntityUtils;
import org.json.JSONException;
import org.json.JSONObject;

public class ClientUtility {
	private static final Log LOG = LogFactory.getLog(ClientUtility.class);
	
	public static final String SESSION_COOKIE_NAME = "session";
	public static final String ORG_COOKIE_NAME = "org";
	public static final String SECRET_KEY_COOKIE_NAME = "api_key";

	public String SERVER_DOMAIN = "";
	private String BASE_URL = "";


	private String orgUid;
	private String sessionUid;
	private String secretKey;
	private int statusCode;
	
	private boolean compress = true;
	
	public boolean isCompress() {
		return compress;
	}

	public void setCompress(boolean compress) {
		this.compress = compress;
	}

	public enum Method {
		PUT, POST, GET, DELETE
	}

	public enum MediaType {
		APPLICATION_JSON
	}

	DefaultHttpClient client = getClient();

	public ClientUtility(String protocol, String domain, String port, String basePath) {
		super();
		SERVER_DOMAIN = domain;
		BASE_URL = protocol + "://" + domain + ":" + port + basePath ;
	}
	
	public int getStatusCode() {
		return statusCode;
	}

	private DefaultHttpClient getClient() {
		DefaultHttpClient client = new DefaultHttpClient();
		client.getParams().setParameter(CoreConnectionPNames.SO_TIMEOUT, (int)120000);
		client.getParams().setParameter(CoreConnectionPNames.CONNECTION_TIMEOUT, (int)12000);
 
		return client;
	}

	public void login(String email, String pwd, String orgUid) {

		try {
			String requestBody = String.format("{\"email\":\"%s\",\"password\":\"%s\"}", email, pwd);
			String result = request(Method.POST, "login", MediaType.APPLICATION_JSON, requestBody);
			JSONObject jresult = new JSONObject(result);
			if (!jresult.getBoolean("success")) {
				System.out.println(jresult.getString("message"));
				return;
			}
			this.sessionUid = jresult.getString("sessionUid");
			this.orgUid = orgUid;

		} catch (JSONException e) {
			throw new RuntimeException(e);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

	}

	public void loginViaSecretKey(String secretKey, String orgUid) {
		this.sessionUid = null;
		this.secretKey = secretKey;
		this.orgUid = orgUid;
	}

/*	public String login() {
	
		try {
			String result = request(Method.POST, "login", MediaType.APPLICATION_JSON,
					"{\"email\":\"a\",\"password\":\"a\"}");
			JSONObject jresult = new JSONObject(result);
			if (!jresult.getBoolean("success")) {
				System.out.println(jresult.getString("message"));
				return "";
			}
			sessionUid = jresult.getString("sessionUid");
			if (jresult.getJSONArray("organization").length() > 0) {
				orgUid = jresult.getJSONArray("organization").getJSONObject(0).getString("uid");
			}
	
		} catch (JSONException e) {
			throw new RuntimeException(e);
		}

		return sessionUid;
	}
*/


	public String getOrgUid() {
		return orgUid;
	}

/*	public void setOrgUid(String orgUid) {
		this.orgUid = orgUid;
		Cookie cookie = new Cookie("localhost", ORG_COOKIE_NAME, orgUid, "/", null, false);
		cookie.setExpiryDate(null);

		client.getState().addCookie(cookie);
	}
*/
	public String getSessionUid() {
		return sessionUid;
	}

	public String request(Method method, String url, MediaType mediaType) {
		return request(method, url, mediaType, "", null);
	}

	public String request(Method method, String url, MediaType mediaType, String body) {
		return request(method, url, mediaType, body, null);
	}

	public String request(Method method, String url, MediaType mediaType, String body, Map<String, Object> params) {
		
		if (sessionUid != null) {
			if (params == null) {
				params = new HashMap<String, Object>();
			}
			params.put(SESSION_COOKIE_NAME, sessionUid);
		} else if (this.secretKey != null) {
			if (params == null) {
				params = new HashMap<String, Object>();
			}
			params.put(SECRET_KEY_COOKIE_NAME, this.secretKey);
		}

		// Assemble uri with query string parameters
		if (params != null) {
			List<NameValuePair> parameters = new ArrayList<NameValuePair>(params.size());
			for (Map.Entry<String, Object> entry : params.entrySet()) {
				parameters.add(new BasicNameValuePair(entry.getKey(), entry.getValue().toString()));
			}
			String qM = url.contains("?") ? "" : "?";
			url = BASE_URL + url + qM + "&" + URLEncodedUtils.format(parameters, Consts.UTF_8);
		} else {
			url = BASE_URL + url;
		}

		HttpRequestBase meth = null;
		if (method == Method.GET) {
			meth = new HttpGet(url);
		} else if (method == Method.PUT) {
			HttpPut putMethod = new HttpPut(url);
			if (compress) {
				byte[] compressed = compress(body.getBytes());
				putMethod.setEntity(new ByteArrayEntity(compressed));
			} else {
				putMethod.setEntity(new StringEntity(body, Consts.UTF_8));
			}
			meth = putMethod;
		} else if (method == Method.POST) {
			HttpPost postMethod = new HttpPost(url);
			if (compress) {
				byte[] compressed = compress(body.getBytes());
				postMethod.setEntity(new ByteArrayEntity(compressed));
			} else {
				postMethod.setEntity(new StringEntity(body, Consts.UTF_8));
			}
			meth = postMethod;
		} else if (method == Method.DELETE) {
			meth = new HttpDelete(url);
		}

		if (orgUid != null) {
			meth.addHeader("Cookie", ORG_COOKIE_NAME + "=" + orgUid);
		}
		
		meth.addHeader("Accept", "application/json");
		meth.addHeader("Content-type", "application/json");
		
		if (compress) {
			LOG.debug("Send compressing");
			meth.addHeader("Content-Encoding", "gzip");			
		}

		try {
			LOG.debug("Send calling executing method");
			HttpResponse response = client.execute(meth);
			LOG.debug("Send executing method complete");
			
			HttpEntity entity = response.getEntity();
			String responseBody = EntityUtils.toString(entity);
			EntityUtils.consume(entity);

			statusCode = response.getStatusLine().getStatusCode();
			if (response.getStatusLine().getStatusCode() == HttpStatus.SC_OK) {
				LOG.debug("Send statusCode == HttpStatus.SC_OK - getting response body");
				return responseBody;

			} else {			
				LOG.debug("Http statusCode = " + response.getStatusLine().getStatusCode() + ": " + responseBody);
				throw new RuntimeException("Http statusCode = " + response.getStatusLine().getStatusCode() + ": " + responseBody);
			}

		} catch (IOException e) {
			LOG.debug("Fatal transport error: " + e.getMessage());
			throw new RuntimeException(e);
		} finally {
			LOG.debug("Send releasing connection");
			// Release the connection.
			meth.releaseConnection();
		}

	}

/*	public String request(Method method, String url, MediaType mediaType, String body, Map<String, Object> params,
			String appKey, String orgUid) {

		HttpMethod meth = null;
		if (method == Method.GET) {
			meth = new GetMethod(BASE_URL + url);
		} else if (method == Method.PUT) {
			PutMethod putMethod = new PutMethod(BASE_URL + url);
			if (compress) {
				byte[] compressed = compress(body.getBytes());
				putMethod.setRequestEntity(   new ByteArrayRequestEntity(compressed, null));
			} else {
				putMethod.setRequestEntity(new StringRequestEntity(body));
			}
			meth = putMethod;
		} else if (method == Method.POST) {
			PostMethod postMethod = new PostMethod(BASE_URL + url);
			if (compress) {
				byte[] compressed = compress(body.getBytes());
				postMethod.setRequestEntity(   new ByteArrayRequestEntity(compressed, null));
			} else {
				postMethod.setRequestEntity(new StringRequestEntity(body));
			}
			meth = postMethod;
		} else if (method == Method.DELETE) {
			meth = new DeleteMethod(BASE_URL + url);
		}

		if (params != null) {
			NameValuePair[] nvp = new NameValuePair[params.size()];
			int i = 0;
			for (Map.Entry<String, Object> entry : params.entrySet()) {
				nvp[i++] = new NameValuePair(entry.getKey(), entry.getValue().toString());
			}
			meth.setQueryString(nvp);
		}

		meth.addRequestHeader("Cookie", ORG_COOKIE_NAME + "=" + orgUid + "; " + SECRET_KEY_COOKIE_NAME + "=" + appKey);
		meth.setRequestHeader("Accept", "application/json");
		meth.setRequestHeader("Content-type", "application/json");

		try {
			LOG.debug("Send executeMethod");
			statusCode = client.executeMethod(meth);
			LOG.debug("Send executeMethod complete");

			if (statusCode == HttpStatus.SC_OK) {

				LOG.debug("Getting response body");
				return meth.getResponseBodyAsString();

			} else {			
				LOG.debug("statusCode != HttpStatus.SC_OK: " + meth.getStatusLine());
				LOG.debug("statusCode != HttpStatus.SC_OK: " + meth.getResponseBodyAsString());

				throw new RuntimeException(meth.getStatusLine().toString());
			}

		} catch (HttpException e) {
			LOG.debug("Fatal protocol violation: " + e.getMessage());
			throw new RuntimeException(e);

		} catch (IOException e) {
			LOG.debug("Fatal transport error: " + e.getMessage());
			throw new RuntimeException(e);
		} finally {
			LOG.debug("Send releasing connection");
			// Release the connection.
			meth.releaseConnection();
		}

	}
	*/
    public static byte[] compress(byte[] content){
        ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
        try{
            GZIPOutputStream gzipOutputStream = new GZIPOutputStream(byteArrayOutputStream);
            gzipOutputStream.write(content);
            gzipOutputStream.close();
        } catch(IOException e){
            throw new RuntimeException(e);
        }
        LOG.debug(String.format("Compressed to %d bytes (%f%% of original size of %d bytes)\n", byteArrayOutputStream.size(), (100.0f * byteArrayOutputStream.size()/content.length), content.length));
        return byteArrayOutputStream.toByteArray();
    }
}
