package com.grovestreams.sensor.gs_gmetad;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TimeZone;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class GangliaXmlConverter {

	public static Map<String, Long> pointUpdateMap = new HashMap<String, Long>();

	public static String convertToJson(long pollingInterval, List<Reading> readings, String lastResult,
			List<String> includes, List<String> excludes) throws JSONException,
			ParserConfigurationException, UnsupportedEncodingException, SAXException, IOException {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder = factory.newDocumentBuilder();

		boolean hasMetricsToUpload = false;

		JSONArray jcomps = new JSONArray();

		JSONArray jmissing = new JSONArray();
		if (lastResult != null && !lastResult.isEmpty()) {
			JSONObject jobj = new JSONObject(lastResult);
			if (jobj.has("missing")) {
				jmissing = jobj.getJSONArray("missing");
			}
		}

		for (Reading reading : readings) {
			Document document = builder.parse(new InputSource(new ByteArrayInputStream(reading.getXml().getBytes(
					"utf-8"))));

			Element xmlGanglia = (Element) document.getElementsByTagName("GANGLIA_XML").item(0);

			NodeList xmlClusters = xmlGanglia.getElementsByTagName("CLUSTER");
			for (int c = 0; c < xmlClusters.getLength(); c++) {
				Element xmlClusterElem = (Element) xmlClusters.item(c);

				String clusterName = xmlClusterElem.getAttribute("NAME");
				if (clusterName.equals("unspecified")) {
					clusterName = "";
				}
				
				String latlong = xmlClusterElem.getAttribute("LATLONG");

				NodeList xmlHosts = xmlClusterElem.getElementsByTagName("HOST");
				for (int h = 0; h < xmlHosts.getLength(); h++) {
					Element xmlHostElem = (Element) xmlHosts.item(h);

					String hostIp = xmlHostElem.getAttribute("IP");
					String hostName = xmlHostElem.getAttribute("NAME");
					String location = xmlHostElem.getAttribute("LOCATION");

					String compName = hostName;

					boolean sendMeta = wasCompMissing(jmissing, compName);
					JSONObject jcomp = getOrCreateComp(compName, clusterName, hostIp, location, latlong, jcomps, xmlHostElem,
							sendMeta);
					
					//Should point stream data be sent (we send it initially and once a day)
					boolean sendPtStreamData = false;
					Calendar cal = new GregorianCalendar(TimeZone.getTimeZone("UTC"));
					long now = cal.getTimeInMillis();
					long span = now - (pointUpdateMap.get(compName) == null ? 0 : pointUpdateMap.get(compName));
					if (sendMeta || span > 86400 * 1000) {
						sendPtStreamData = true;
						pointUpdateMap.put(compName, now);
					}

					NodeList xmlMetrics = xmlHostElem.getElementsByTagName("METRIC");
					for (int m = 0; m < xmlMetrics.getLength(); m++) {
						Element xmlMetricElem = (Element) xmlMetrics.item(m);

						String id = xmlMetricElem.getAttribute("NAME");

						if (includeItem(id, includes, excludes)) {
							String desc = "";
							String name = "";
							String group = "";
							String type = xmlMetricElem.getAttribute("TYPE");
							String unit = xmlMetricElem.getAttribute("UNITS");
							String slope = xmlMetricElem.getAttribute("SLOPE");
							String val = xmlMetricElem.getAttribute("VAL");
							String tn = xmlMetricElem.getAttribute("TN");

							NodeList xmlEDs = xmlMetricElem.getElementsByTagName("EXTRA_DATA");
							if (xmlEDs.getLength() > 0) {
								Element xmlED = (Element) xmlEDs.item(0);
								NodeList xmlEEs = xmlED.getElementsByTagName("EXTRA_ELEMENT");
								for (int n = 0; n < xmlEEs.getLength(); n++) {
									Element xmlEE = (Element) xmlEEs.item(n);
									String eeName = xmlEE.getAttribute("NAME");
									String eeVal = xmlEE.getAttribute("VAL");
									if (eeName.equals("GROUP")) {
										group = eeVal;
									} else if (eeName.equals("TITLE")) {
										name = eeVal;
									} else if (eeName.equals("DESC")) {
										desc = eeVal;
									}
								}
							}
							if (name.isEmpty()) {
								name = id;
							}

							hasMetricsToUpload = true;
							sendMeta = wasCompItemMissing(jmissing, compName, "stream", id);
							if (isTimeSeriesStream(slope)) {
								// Stream (Intvl or Rdm)
								
								setStreamData(pollingInterval, reading.getDate(), id, name, desc, group, type, unit,
										slope, tn, val, jcomp, sendMeta);
								
							} else {
								//point_stream - only update initially and once a day
							
								if (sendMeta || sendPtStreamData) {
									
									setStreamData(pollingInterval, reading.getDate(), id, name, desc, group, type, unit,
											slope, tn, val, jcomp, sendMeta);
									
								}
								
							}

						}

					}
				}
			}
		}

		if (hasMetricsToUpload) {

			JSONObject jfeed = new JSONObject();
			jfeed.put("component", jcomps);

			JSONObject jobj = new JSONObject();
			jobj.put("feed", jfeed);

			return jobj.toString();
		} else {
			return "";
		}

	}

	private static boolean includeItem(String id, List<String> includes, List<String> excludes) {
		//If include is a substring then include it
		//If exclude is a substring then exclude it
		for (String filter : includes) {
			if (id.contains(filter)){
				return true;
			}
		}
		for (String filter : excludes) {
			if (id.contains(filter)){
				return false;
			}
		}
		
		return true;
	}

	private static boolean isTimeSeriesStream(String slope) {
		if (slope.equals("both")) {
			return true;
		}
		return false;
	}

	private static boolean wasCompMissing(JSONArray jmissing, String compId) throws JSONException {
		for (int i = 0; i < jmissing.length(); i++) {
			JSONObject jitem = jmissing.getJSONObject(i);
			if (jitem.has("componentId") && jitem.getString("componentId").equals(compId) && !jitem.has("itemId")) {
				return true;
			}
		}
		return false;
	}

	private static boolean wasCompItemMissing(JSONArray jmissing, String compId, String itemType, String charId)
			throws JSONException {
		for (int i = 0; i < jmissing.length(); i++) {
			JSONObject jitem = jmissing.getJSONObject(i);
			if (jitem.has("componentId") && jitem.has("itemId") && jitem.getString("componentId").equals(compId)
					&& jitem.getString("itemType").equals(itemType) && jitem.getString("itemId").equals(charId)) {
				return true;
			}
		}
		return false;
	}
	
	private static boolean isDecimalType(String type) {
		if (type.equals("float") || type.equals("double")) {
			return true;
		} else {
			return false;
		} 
	}

	private static String getValueType(String type) {
		// string | int8 | uint8 | int16 | uint16 | int32 | uint32 | float |
		// double | timestamp

		if (type.equals("int8") || type.equals("uint8") || type.equals("int16") || type.equals("uint16")) {
			return "SHORT";
		} else if (type.equals("int32") || type.equals("uint32")) {
			return "INTEGER";
		} else if (type.equals("timestamp")) {
			return "LONG";
		} else if (type.equals("float")) {
			return "FLOAT";
		} else if (type.equals("double")) {
			return "DOUBLE";
		} else if (type.equals("string")) {
			return "STRING";
		}
		return "";
	}

	private static String getUnitId(String units, String type, String streamId) {

		if (streamId.contains("boottime")) {
				return "gs.days";
		} else if (streamId.contains("_avg_time")) {
			return "gs.milliseconds";
		} else if (units.equals("KB")) {
			return "gs.KB";
		} else if (units.equals("%")) {
			return "gs.percent";
		} else if (units.equals("s")) {
			return "gs.seconds";
		} else if (units.equals("MHz")) {
			return "gs.mhz";
		} else if (units.equals("bytes/sec")) {
			return "gs.bytesPerSecond";
		} else if (units.equals("CPUs")) {
			return "gs.cpus";
		} else if (units.equals("GB")) {
			return "gs.GB";
		} else if (units.equals("packets/sec")) {
			return "gs.packetsPerSecond";
		} 

		if (isDecimalType(type)) {
			return "gs.noSymbolDec";
		} else {
			//Need this for bools even if it is not a number
			return "gs.noSymbol";
		}

	}

	private static String getCycleId(String streamId) {

		if (streamId.equals("cpu_user") || streamId.equals("cpu_system") || streamId.equals("cpu_idle")
				|| streamId.equals("cpu_nice") || streamId.equals("cpu_aidle") || streamId.equals("cpu_wio")
				|| streamId.equals("load_one") || streamId.equals("load_five") || streamId.equals("load_fifteen")
				|| streamId.equals("proc_run") || streamId.equals("proc_total")) {

			return "gs.20sec";
		} else if (streamId.equals("mem_free") || streamId.equals("mem_shared") || streamId.equals("mem_buffers")
				|| streamId.equals("mem_cached") || streamId.equals("swap_free") || streamId.equals("bytes_out")
				|| streamId.equals("bytes_in") || streamId.equals("pkts_in") || streamId.equals("pkts_out")
				|| streamId.equals("disk_free") || streamId.equals("part_max_used")) {

			return "gs.minutes";
		}

		return "gs.10Sec";
	}

	private static String getRcId(String streamId) {

		if (streamId.equals("cpu_user") || streamId.equals("cpu_system") || streamId.equals("cpu_idle")
				|| streamId.equals("cpu_nice") || streamId.equals("cpu_aidle") || streamId.equals("cpu_wio")
				|| streamId.equals("load_one") || streamId.equals("load_five") || streamId.equals("load_fifteen")
				|| streamId.equals("proc_run") || streamId.equals("proc_total") || streamId.equals("mem_free")
				|| streamId.equals("mem_shared") || streamId.equals("mem_buffers") || streamId.equals("mem_cached")
				|| streamId.equals("swap_free") || streamId.equals("bytes_out") || streamId.equals("bytes_in")
				|| streamId.equals("pkts_in") || streamId.equals("pkts_out") || streamId.equals("disk_free")
				|| streamId.equals("disk_total") || streamId.equals("part_max_used")

		) {
			//Associate with a rollup calendar
			return "gs.YQMDH5M";
		}

		return null;
	}


	private static void setStreamData(long pollingInterval, long readingDate, String streamId, String streamName,
			String streamDesc, String streamGroup, String type, String unit, String slope, String tn, String val,
			JSONObject jcomp, boolean sendMeta) throws JSONException {

		JSONArray jstreams = null;
		if (jcomp.has("stream")) {
			jstreams = jcomp.getJSONArray("stream");
		} else {
			jstreams = new JSONArray();
			jcomp.put("stream", jstreams);
		}

		JSONObject jstream = getOrCreateStream(jstreams, streamId, streamName, streamDesc, streamGroup, type, unit,
				slope, sendMeta);

		JSONArray jdata = jstream.getJSONArray("data");
		JSONArray jtime = jstream.getJSONArray("time");

		long actualMetricDate = readingDate - Long.valueOf(tn) * 1000;
		jtime.put(actualMetricDate);
		
	    if (streamId.equals("boottime")) {
	    	Calendar cal = new GregorianCalendar(TimeZone.getTimeZone("UTC"));
			long upTimeMillis = cal.getTimeInMillis() - Long.valueOf(val)*1000;
			
	    	//Custom adjustment
	    	Double dval = upTimeMillis / 86400000d;
	    	val = Integer.toString(dval.intValue());
	    }
	
		jdata.put(val);
	}

	private static JSONObject getOrCreateStream(JSONArray jstreams, String streamId, String streamName,
			String streamDesc, String streamGroup, String type, String unit, String slope, boolean sendMeta)
			throws JSONException {

		JSONObject jstream = null;
		for (int i = 0; i < jstreams.length(); i++) {
			if (jstreams.getJSONObject(i).getString("streamId").equals(streamId)) {
				jstream = jstreams.getJSONObject(i);
				break;
			}
		}
		if (jstream == null) {
			// Create one and setup auto-creation defaults. Adding default
			// information will allow GroveStreams
			// to create the stream as the feed is uploaded if it doesn't
			// already exist for the component or uses the
			// component template if the template has a stream with the same ID.
			jstream = new JSONObject();
			jstreams.put(jstream);

			jstream.put("streamId", streamId);

			jstream.put("data", new JSONArray());
			jstream.put("time", new JSONArray());

			if (sendMeta) {
				JSONObject jdefaults = new JSONObject();
				jstream.put("defaults", jdefaults);

				jdefaults.put("name", streamName);
				jdefaults.put("description", streamDesc);

				jdefaults.put("valueType", getValueType(type));
				jdefaults.put("baseCycleId", getCycleId(streamId));
				jdefaults.put("unitId", getUnitId(unit, type, streamId));
				if (isTimeSeriesStream(slope)) {
					String rcId = getRcId(streamId);
					if (rcId != null) {
						jdefaults.put("rollupCalendarId", rcId);
					}					
					jdefaults.put("streamType", "intvl_stream");
					// Gap Filling
					JSONObject jgapfilling = new JSONObject();
					jgapfilling.put("gapFillType", "PREVIOUS_VAL");
					jgapfilling.put("maxGapsToFill", 20);
					jdefaults.put("gap_filling", jgapfilling);
					
				} else {
					jdefaults.put("streamType", "point_stream");
				}

				//Group
				if (!streamGroup.isEmpty()) {
					JSONObject jgroup = new JSONObject();
					jgroup.put("name", streamGroup);
					jgroup.put("id", streamGroup);
					jdefaults.put("comp_group", jgroup);
				}

				// Visuals
				if (unit.equals("%")) {
					JSONObject jvisuals = new JSONObject();
					jvisuals.put("hasAxisMinMax", "true");
					jvisuals.put("chartAxisMax", "100");
					jdefaults.put("visuals", jvisuals);
				}
			}

		}

		return jstream;
	}

	private static JSONObject getOrCreateComp(String name, String clusterName, String hostIp, String location, String latlong,
			JSONArray jcomps, Element xmlHostElem, boolean sendMeta) throws JSONException {
		for (int i = 0; i < jcomps.length(); i++) {
			if (jcomps.getJSONObject(i).getString("componentId").equals(name)) {
				return jcomps.getJSONObject(i);
			}
		}

		// Setting the id and template name will trigger the server to create a
		// new component if one doesn't exist based on the template
		JSONObject jcomp = new JSONObject();
		jcomp.put("componentId", name); //IP address seems to change when there are multiple nics so use the fqn as the id.

		if (sendMeta) {
			jcomp.put("componentTemplateId", "gs.ganglia");
			JSONObject jCompDefaults = new JSONObject();
			jCompDefaults.put("name", name);
			
			String description = "IP At Registration: " + hostIp;
			if (!clusterName.isEmpty()) {
				description += "<br>Cluster: " + clusterName;
			}
			jCompDefaults.put("description", description);
			
			if (!location.isEmpty() && !location.equals("unspecified")) {
				jCompDefaults.put("folderPath", location);
			}

			JSONObject jcompLoc = new JSONObject();
			
			jcompLoc.put("description", location);
			if (!latlong.isEmpty()) {
				Double[] dblLatLong = GeoConverter.computeLatLonFromString(latlong);
				if (dblLatLong != null) {
					jcompLoc.put("latitude", dblLatLong[0]);
					jcompLoc.put("longitude", dblLatLong[1]);
				}
			}

			jCompDefaults.put("location", jcompLoc);
			jcomp.put("defaults", jCompDefaults);
		}

		jcomps.put(jcomp);

		return jcomp;
	}

	private static boolean metricExists(Element xmlHostElem, String metricName) {
		NodeList xmlMetrics = xmlHostElem.getElementsByTagName("METRIC");
		for (int m = 0; m < xmlMetrics.getLength(); m++) {
			Element xmlMetricElem = (Element) xmlMetrics.item(m);
			String metricUploading = xmlMetricElem.getAttribute("NAME");
			// System.out.println(metricUploading+ ": " +
			// metricUploading.indexOf(metricName));
			if (metricUploading.indexOf(metricName) >= 0) {
				return true;
			}
		}

		return false;
	}

}
