package com.grovestreams.sensor.gs_gmetad;

import java.util.regex.Matcher;
import java.util.regex.Pattern;


//Pulled this from the web - totally untested but seems to work as long as N,E,W,S are at the end, not at the beginning.
public class GeoConverter {
	

	public static Double[] computeLatLonFromString(String coordString)
    {
		
        try {
			if (coordString == null)
			{
			    throw new IllegalArgumentException();
			}

			Double lat = null;
			Double lon = null;
			coordString = coordString.trim();
			String regex;
			String separators = "(\\s*|,|,\\s*)";
			Pattern pattern;
			Matcher matcher;
			
			// Try to extract a pair of signed decimal values separated by a space, ',' or ', '
			// Allow E, W, S, N sufixes
			if (lat == null || lon == null)
			{
			    regex = "([-|\\+]?\\d+?(\\.\\d+?)??\\s*[N|n|S|s]??)";
			    regex += separators;
			    regex += "([-|\\+]?\\d+?(\\.\\d+?)??\\s*[E|e|W|w]??)";
			    pattern =  Pattern.compile(regex);
			    matcher = pattern.matcher(coordString);
			    if (matcher.matches())
			    {
			        String sLat = matcher.group(1).trim();  // Latitude
			        int signLat = 1;
			        char suffix = sLat.toUpperCase().charAt(sLat.length() - 1);
			        if (!Character.isDigit(suffix))
			        {
			            signLat = suffix == 'N' ? 1 : -1;
			            sLat = sLat.substring(0, sLat.length() - 1);
			            sLat = sLat.trim();
			        }

			        String sLon = matcher.group(4).trim();  // Longitude
			        int signLon = 1;
			        suffix = sLon.toUpperCase().charAt(sLon.length() - 1);
			        if (!Character.isDigit(suffix))
			        {
			            signLon = suffix == 'E' ? 1 : -1;
			            sLon = sLon.substring(0, sLon.length() - 1);
			            sLon = sLon.trim();
			        }

			        lat = Double.parseDouble(sLat) * signLat;
			        lon = Double.parseDouble(sLon) * signLon;
			    }
			}

			// Try to extract two degrees minute seconds blocks separated by a space, ',' or ', '
			// Allow S, N, W, E suffixes and signs.
			// eg: -123� 34' 42" +45� 12' 30"
			// eg: 123� 34' 42"S 45� 12' 30"W
			if (lat == null || lon == null)
			{
			    regex = "([-|\\+]?\\d{1,3}[d|D|\u00B0|\\s](\\s*\\d{1,2}['|\u2019|\\s])?(\\s*\\d{1,2}[\"|\u201d])?\\s*[N|n|S|s]?)";
			    regex += separators;
			    regex += "([-|\\+]?\\d{1,3}[d|D|\u00B0|\\s](\\s*\\d{1,2}['|\u2019|\\s])?(\\s*\\d{1,2}[\"|\u201d])?\\s*[E|e|W|w]?)";
			    pattern =  Pattern.compile(regex);
			    matcher = pattern.matcher(coordString);
			    if (matcher.matches())
			    {
			        lat = parseDMSString(matcher.group(1));
			        lon = parseDMSString(matcher.group(5));
			    }
			}
			if (lat == null || lon == null)
			{
			    regex = "([-|\\+]?\\d{1,3}[d|D|\u00B0|\\s](\\s*\\d{1,2}['|\u2019|\\s])?(\\s*\\d{1,2}[\"|\u201d])?\\s*[N|n|S|s]?)";
			    regex += separators;
			    regex += "([-|\\+]?\\d{1,3}[d|D|\u00B0|\\s](\\s*\\d{1,2}['|\u2019|\\s])?(\\s*\\d{1,2}[\"|\u201d])?\\s*[E|e|W|w]?)";
			    pattern =  Pattern.compile(regex);
			    matcher = pattern.matcher(coordString);
			    if (matcher.matches())
			    {
			        lat = parseDMSString(matcher.group(1));
			        lon = parseDMSString(matcher.group(5));
			    }
			}


			if (lat == null || lon == null)
			    return null;

			if(lat >= -90 && lat <= 90 && lon >= -180 && lon <= 180) {
				Double[] latlon = new Double[2];
				latlon[0] = lat;
				latlon[1] = lon;
			    return latlon;
			}
		} catch (NumberFormatException e) {
			//Do nothing - return null
		}

        return null;
    }
	
    private static Double parseDMSString(String dmsString)
    {
        // Replace degree, min and sec signs with space
        dmsString = dmsString.replaceAll("[D|d|\u00B0|'|\u2019|\"|\u201d]", " ");
        // Replace multiple spaces with single ones
        dmsString = dmsString.replaceAll("\\s+", " ");
        dmsString = dmsString.trim();

        // Check for sign prefix and suffix
        int sign = 1;
        char suffix = dmsString.toUpperCase().charAt(dmsString.length() - 1);
        if (!Character.isDigit(suffix))
        {
            sign = (suffix == 'N' || suffix == 'E') ? 1 : -1;
            dmsString = dmsString.substring(0, dmsString.length() - 1);
            dmsString = dmsString.trim();
        }
        char prefix = dmsString.charAt(0);
        if (!Character.isDigit(prefix))
        {
            sign *= (prefix == '-') ? -1 : 1;
            dmsString = dmsString.substring(1, dmsString.length());
        }

        // Process degrees, minutes and seconds
        String[] DMS = dmsString.split(" ");
        double d = Integer.parseInt(DMS[0]);
        double m = DMS.length > 1 ? Integer.parseInt(DMS[1]) : 0;
        double s = DMS.length > 2 ? Integer.parseInt(DMS[2]) : 0;

        if (m >= 0 && m <= 60 && s >= 0 && s <= 60)
            return d * sign + m / 60 * sign + s / 3600 * sign;
        
        return null;
    }
	
}
