package com.grovestreams.sensor.gs_gmetad;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Properties;

import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class GsFeeder {
	private static final Log LOG = LogFactory.getLog(GsFeeder.class);

	private static final ScheduledExecutorService scheduler = Executors.newScheduledThreadPool(3);

	GsFeeder() {
	}

	public static void main(String[] args) {
		LOG.info("gs_gmetad started");

		Thread runtimeHookThread = new Thread() {
			public void run() {
				shutdownHook();
			}
		};
		Runtime.getRuntime().addShutdownHook(runtimeHookThread);
		try {
			start();
		      while (true) {
		          Thread.sleep (3000);
		      }
			
		
		} catch (Throwable t) {
			LOG.error("Exception!", t);
		}
	}

	private static void shutdownHook() {
		LOG.info("ShutdownHook started");

		try {
			stop();
		} catch (Exception e) {
			LOG.error("Exception!", e);
		}
		LOG.info("ShutdownHook completed");
	}


	public static void stop() {
		scheduler.shutdown();
	}

	public static void start() throws FileNotFoundException, IOException {

		final String PROPERTY_FILE_NAME = "/gs_gmetad.properties";

		Properties properties = new Properties();
		Queue queue = new Queue(properties);

		InputStream is = GsFeeder.class.getResourceAsStream(PROPERTY_FILE_NAME);
		if (is != null) {
			properties.load(is);
		} else {
			properties.load(new FileInputStream(PROPERTY_FILE_NAME));
		}

		long pollingInterval = Long.valueOf(properties.getProperty("POLL_INTERVAL", "20").trim()) * 1000;
		LOG.info("Polling Interval: " + pollingInterval/1000 + " seconds");
		
		//Schedule the pulling of metrics from Ganglia
		scheduler.scheduleAtFixedRate(new Pusher(properties, queue), 0, pollingInterval, TimeUnit.MILLISECONDS);
		
		//Schedule the pushing of metrics to GroveStreams
		// Push ten seconds after successful upload since we're limited to one call every 10 seconds
		scheduler.scheduleWithFixedDelay(new Popper(properties, queue), 2000, pollingInterval, TimeUnit.MILLISECONDS);
	}

}
