package com.grovestreams.sensor.gs_gmetad;

import java.util.ArrayList;
import java.util.List;
import java.util.Properties;
import java.util.TimerTask;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.http.HttpStatus;

import com.grovestreams.sensor.gs_gmetad.ClientUtility.MediaType;
import com.grovestreams.sensor.gs_gmetad.ClientUtility.Method;

public class Popper extends TimerTask {
	private static final Log LOG = LogFactory.getLog(Popper.class);

	Queue queue;
	Properties properties;
	List<String> excludes = new ArrayList<String>();
	List<String> includes = new ArrayList<String>();
	boolean logged = false; // used to log an error once (or log fills up)
	ClientUtility clientUtility = null;
	String lastResult;

	public Popper(Properties props, Queue queue) {
		super();
		this.queue = queue;
		this.properties = props;
		
		String[] includeMetrics = properties.getProperty("INCLUDE_METRICS", "").split(",");
		for (String includeMetric : includeMetrics) {
			includes.add(includeMetric.trim());
		}
		String[] excludeMetrics = properties.getProperty("EXCLUDE_METRICS", "").split(",");
		for (String excludeMetric : excludeMetrics) {
			excludes.add(excludeMetric.trim());
		}	

		LOG.info("Popper running");
	}

	@Override
	public void run() {
		List<Reading> readings = null;
		try {

			long pollingInterval = Long.valueOf(properties.getProperty("POLL_INTERVAL", "20").trim()) * 1000;
			int maxUploadItems = Integer.valueOf(properties.getProperty("MAX_UPLOAD_ITEMS", "1000").trim());
			readings = queue.getMany(maxUploadItems);

			LOG.info(String.format("Reading %d items from queue", readings.size()));

			String json = GangliaXmlConverter.convertToJson(pollingInterval, readings, lastResult, includes, excludes);

			if (!json.isEmpty()) {
				
				lastResult = sendReadings(json);
			}

			// Pop off the successfully uploaded metrics
			LOG.info(String.format("Popping %d items off of queue", readings.size()));
			queue.remove(readings);

			logged = false;
		} catch (Exception e) {
			// Error occurred - leave items in queue for next upload
			LOG.error("Exception!", e);
			GangliaXmlConverter.pointUpdateMap.clear();
			try {
				if (getClientUtility().getStatusCode() == HttpStatus.SC_FORBIDDEN ||
					getClientUtility().getStatusCode() == HttpStatus.SC_UNPROCESSABLE_ENTITY   ||
					getClientUtility().getStatusCode() == HttpStatus.SC_BAD_REQUEST) {
					//Clear the queue since there may be some bogus entries
					queue.remove(readings);
				}
			} catch (Exception e1) {
				LOG.error("Exception!", e);
			}


			//Reset connection
			
			clientUtility = null;

		}

	}

	private String sendReadings(String json) throws Exception {
		ClientUtility cu = getClientUtility();

		LOG.info("Sending JSON...");
		LOG.debug("JSON: " + json);
		long startTimer = System.currentTimeMillis(); 
		
		//Make the HTTP call to GroveStreams to upload metrics
		String result = cu.request(Method.PUT, "feed?createDefault=false", MediaType.APPLICATION_JSON, json);
		
		long span = System.currentTimeMillis() - startTimer;
		LOG.info("Send result (" + span / 1000.0f + "s): " + result);

		if (span / 1000.0f > 30) {
			LOG.info("Long Send Time!");
		}

		return result;
	}

	private ClientUtility getClientUtility() throws Exception {
		if (clientUtility == null) {
			
			String protocol = properties.getProperty("UPLOAD_PROTOCOL", "").trim();
			String domain = properties.getProperty("UPLOAD_DOMAIN", "").trim();
			String port = properties.getProperty("UPLOAD_PORT", "").trim();
			String basePath = properties.getProperty("UPLOAD_BASE_PATH", "").trim();

			clientUtility = new ClientUtility(protocol, domain, port, basePath);

			String orgUid = properties.getProperty("ORGANIZATION_UID", "").trim();

			String secretKey = properties.getProperty("GROVE_STREAMS_API_KEY", "").trim();

			String email = properties.getProperty("GROVE_STREAMS_USER_EMAIL", "").trim();
			String pwd = properties.getProperty("GROVE_STREAMS_USER_PWD", "").trim();

			if (!secretKey.isEmpty()) {
				clientUtility.loginViaSecretKey(secretKey, orgUid);
			} else {
				clientUtility.login(email, pwd, orgUid);
			}
		}

		return clientUtility;
	}

}
