package com.grovestreams.sensor.gs_gmetad;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Calendar;
import java.util.Properties;
import java.util.TimerTask;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class Pusher extends TimerTask  {
	private static final Log LOG = LogFactory.getLog(Pusher.class);
	Queue queue;
	Properties properties;
	boolean socketLogged = false; //used to log an error once (or log fills up)
	boolean pushLogged = false; //used to log an error once (or log fills up)
	boolean socketOpen = false;
	int bufferLast = 5000; //Used to improve memory allocation for large xml strings


	public Pusher(Properties props, Queue queue) {
		super();
		this.queue = queue;
		this.properties = props;
		LOG.info("Pusher running");
		
	}

	@Override
	public boolean cancel() {
		boolean retVal = super.cancel();
		while (socketOpen) {
			
			try {
				Thread.sleep(100);
			} catch (InterruptedException e) {

			}
		}
		
		return retVal;
	}

	@Override
	public void run() {
		
		
		try {
			String xml;
			try {
				LOG.info("Pusher Getting Xml from ganglia...");
				xml = getMetricsXml();
				socketLogged = false;
			} catch (Exception e) {
				if (!socketLogged || LOG.isInfoEnabled()) {
					LOG.error("Exception!", e);
					socketLogged = true;
				}
				//Bad Socket? Reset it and try again
				xml = getMetricsXml();
			}
			
			LOG.info("Pushing Xml into queue (" + xml.length() + " chars)...");
			LOG.debug("XML: " + xml);			
			
//			writeStringToFile("/home/xxx/ganglia.xml", xml);
			
			queue.push(new Reading(Calendar.getInstance().getTimeInMillis(), xml));
			
			LOG.info("Items in queue: " + queue.size());
			pushLogged= false;
			
		} catch (Exception e) {
			if (!pushLogged || LOG.isInfoEnabled()) {
				LOG.error("Exception!", e);
				pushLogged = true;
			}
			socketOpen = false;
		}
	}
	
	private void writeStringToFile(String fileName, String contents) {
		BufferedWriter writer = null;
		try
		{
			writer = new BufferedWriter( new FileWriter( fileName));
			writer.write( contents);

		}
		catch ( IOException e)
		{
		}
		finally
		{
			try
			{
				if ( writer != null)
					writer.close( );
			}
			catch ( IOException e)
			{
			}
	     }
	}
	
	private String getMetricsXml() throws UnknownHostException, IOException {
		//Get Metrics XML from Ganglia gmond
			
		socketOpen = true;
		LOG.debug("Pusher Getting Socket...");
		Socket socket = getSocket();
		
		LOG.debug("Pusher BufferedReader...");

		//Get Metrics from host
		BufferedReader rd = new BufferedReader(new InputStreamReader(
				socket.getInputStream()));
		
		long lines = 0;
		String str;
		StringBuilder sb = new StringBuilder(bufferLast);
		while ((str = rd.readLine()) != null) {
			// System.out.println(str);
			sb.append(str);
			lines++;
		}
		
		LOG.debug("Pusher BufferedReader closing..." + lines + " lines");
		rd.close();
		socketOpen = false;
		
		bufferLast = (int) (sb.length() + sb.length()*0.05); //add 5%
		bufferLast = Math.min(bufferLast, 20971520); //20 MBs. Used to prevent one large read resulting in a large buffer for every call

		// System.out.println(xml);

		return sb.toString();
		
	}
	
	private Socket getSocket() throws UnknownHostException, IOException {
		String fromHost = properties.getProperty("RETRIEVE_FROM_HOST", "localhost").trim();
	
		int fromHostPort = Integer.valueOf(properties.getProperty("RETRIEVE_FROM_HOST_PORT", "8649").trim());

		LOG.debug("Getting Socket for host: " + fromHost + ":" + fromHostPort);			
		return new Socket(fromHost, fromHostPort);
	}
	

}
