package com.grovestreams.sensor.gs_gmetad;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

public class Queue {
	// FIFO queue - Used to hold metrics from Ganglia until a successfull upload occurs. 
	// Note that the queue is an in memory queue

	private int maxQueueItems = 0;
	private List<Reading> readings = new LinkedList<Reading>();

	public Queue(Properties properties) {
		super();
		this.maxQueueItems = Integer.valueOf(properties.getProperty("MAX_QUEUE_ITEMS", "4320").trim());
	}

	public void push(Reading reading) {
		synchronized (this) {

			if (readings.size() >= maxQueueItems) {
				// Queue has reached its maximum size. Throw out earliest
				int removeCount = readings.size() - maxQueueItems + 1;
				for (int i = 0; i < removeCount; i++) {
					readings.remove(0);
				}
			}

			readings.add(reading);
		}
	}


	public List<Reading> getMany(int maxToPop) {
		
		//Get all or the max allowed
		List<Reading> readingsToReturn = new ArrayList<Reading>();

		synchronized (this) {

			for (int i=0; i<maxToPop && i<readings.size(); i++) {
				readingsToReturn.add(readings.get(i));
			}

			return readingsToReturn;
		}
	}
	
	public void remove(List<Reading> readingsToRemove) {
		synchronized (this) {
			for (Reading reading : readingsToRemove) {
				try {
					readings.remove(reading);
				} catch (Exception e) {
					// A reading may have been popped off if the queue exceeded
					//  maximum size so just catch the error and continue
				}
			}
			
		}
	}
	
	public int size() {
		return readings.size();
	}

}
